
package rbas.ranksearch.sem;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.util.MatrixUtils;

import java.io.Serializable;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Random;

/**
 * EM estimator for a finite mixture of SEMs with latent variables
 *
 * @author Ricardo Silva
 * @version $Revision: 1.39 $ $Date: 2004/03/26 07:09:22 $
 */

public class MixtureOfSemsEstimator implements Serializable {


    Graph   graph, originalGraph;
    RectangularDataSet dataSet;
    int     numLatents, numVars, sampleSize;

    double  rawData[][];
    Node    latents[], indicators[];
    int     xxParents[][], xyParents[][], yyParents[][], yxParents[][];

    double pi[];                      //mixture probabilities
    double h[][];                     //expected mixture probabilities
    double eX[][][], eXX[][][][];     //marginal implied moments of hidden variables (X variables)
    double beta[][][];                //coefficients into hidden variables
    double upsilon[][];               //latent variance
    double lambda[][][];              //coefficients into observed variables (has one extra entry for intercepts)
    double invPsi[];                  //error term variances for each variables (the same for all mixtures)

    double impliedMu[][];

    double score = -Double.MAX_VALUE;

    int numMixtures;

    protected static int    MAX_ITER = 100;
    protected static double MIN_TOLERANCE = 1.e-5;

    protected boolean converging = true;

    protected int numParameters;

    double impliedSecondMoments[][][];
    double impliedCovarObserved[][][], invImplCovarObs[][][], detCov[];

    public MixtureOfSemsEstimator(RectangularDataSet dataSet, Graph graph, int numMixtures) {
        this.dataSet = dataSet;
        this.graph = new EdgeListGraph(graph);
        this.originalGraph = new EdgeListGraph(graph);
        this.numMixtures = numMixtures;
        initialize();
    }

    private void initialize() {
        //Graph
        List measures = new ArrayList();
        List latentsList = new ArrayList();
        for (Iterator it = this.graph.getNodes().iterator(); it.hasNext();) {
            Node next = (Node) it.next();
            if (next.getNodeType() == NodeType.MEASURED)
                measures.add(next);
            else if (next.getNodeType() == NodeType.LATENT)
                latentsList.add(next);
        }

        this.numVars = measures.size();
        this.numLatents = latentsList.size();
        this.latents = new Node[this.numLatents];
        for (int i = 0; i < this.numLatents; i++)
            this.latents[i] = (Node) latentsList.get(i);

        this.numParameters = this.numVars + (this.numMixtures - 1);
        this.indicators = new Node[this.numVars];
        this.yxParents = new int[this.numVars][];
        for (int i = 0; i < this.numVars; i++) {
            this.indicators[i] = (Node) measures.get(i);
            List xP = new ArrayList();
            for (Iterator it = this.graph.getParents(this.indicators[i]).iterator(); it.hasNext();) {
                Node nextParent = (Node) it.next();
                if (nextParent.getNodeType() == NodeType.LATENT)
                    xP.add(nextParent);
            }

            this.yxParents[i] = new int[xP.size()];
            this.numParameters += xP.size() * this.numMixtures;
            int countPx = 0;
            for (Iterator it = xP.iterator(); it.hasNext();) {
                Object next = it.next();
                if (next instanceof Integer)
                    this.yxParents[i][countPx++] = this.numLatents + ((Integer) next).intValue();
                else
                    for (int j = 0; j < this.latents.length; j++)
                        if (this.latents[j].toString().equals(next.toString())) {
                            this.yxParents[i][countPx++] = j;
                            break;
                        }
            }
        }
        this.yyParents = new int[this.numVars][];
        for (int i = 0; i < this.numVars; i++) {
            List yP = new ArrayList();
            for (Iterator it = this.graph.getParents(this.indicators[i]).iterator(); it.hasNext();) {
                Node nextParent = (Node) it.next();
                if (nextParent.getNodeType() == NodeType.MEASURED)
                    yP.add(nextParent);
            }
            this.yyParents[i] = new int[yP.size() + 1];
            this.numParameters += (yP.size() + 1) * this.numMixtures;
            int countP = 0;
            for (Iterator it = yP.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                for (int j = 0; j < this.indicators.length; j++)
                    if (this.indicators[j].toString().equals(next.toString())) {
                        this.yyParents[i][countP++] = j;
                        break;
                    }
            }
            this.yyParents[i][yP.size()] = this.numVars; //this will correspond to the intercept (mu) term
        }

        this.xxParents = new int[this.numLatents][];
        this.xyParents = new int[this.numLatents][];
        for (int i = 0; i < this.numLatents; i++) {
            List xP = this.graph.getParents(this.latents[i]);
            List xPX = new ArrayList();
            List xPY = new ArrayList();
            for (Iterator it = xP.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                if (next.getNodeType() == NodeType.LATENT)
                    xPX.add(next);
                else
                    xPY.add(next);
            }
            this.xyParents[i] = new int[xPY.size()];
            this.numParameters += xP.size() * this.numMixtures;
            int countPy = 0;
            for (Iterator it = xPY.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                for (int j = 0; j < this.indicators.length; j++)
                    if (this.indicators[j].toString().equals(next.toString())) {
                        this.xyParents[i][countPy++] = j;
                        break;
                    }
            }
            this.xxParents[i] = new int[xPX.size()];
            this.numParameters += xPX.size() * this.numMixtures;
            int countP = 0;
            for (Iterator it = xPX.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                for (int j = 0; j < this.latents.length; j++)
                    if (this.latents[j].toString().equals(next.toString())) {
                        this.xxParents[i][countP++] = j;
                        break;
                    }
            }
        }

        //Data
        this.sampleSize = this.dataSet.getNumRows();
        this.rawData = new double[this.sampleSize][this.numVars + 1];
        for (int i = 0; i < this.numVars; i++) {
            Node v = this.dataSet.getVariable(this.indicators[i].toString());
            double raw0[] = this.dataSet.getDoubleMatrix().viewColumn(this.dataSet.getColumn(v)).toArray();
            for (int j = 0; j < raw0.length; j++)
                this.rawData[j][i] = raw0[j];
        }
        for (int j = 0; j < this.sampleSize; j++)
            this.rawData[j][this.numVars] = 1.;

        //Parameters for Pi
        this.pi = new double[this.numMixtures];
        //Parameters for S
        this.h = new double[this.numMixtures][this.sampleSize];

        //Parameters for X, Z
        this.beta = new double[this.numMixtures][this.numLatents][];
        for (int d = 0; d < this.numMixtures; d++)
            for (int x = 0; x < this.numLatents; x++) {
                int numParents = this.xxParents[x].length + this.xyParents[x].length;
                this.beta[d][x] = new double[numParents];
            }

        this.upsilon = new double[this.numMixtures][this.numLatents];

        this.eX = new double[this.numMixtures][this.sampleSize][this.numLatents];
        this.eXX = new double[this.numMixtures][this.sampleSize][this.numLatents][this.numLatents];

        //Parameters for Y

        this.lambda = new double[this.numMixtures][this.numVars][];
        for (int d = 0; d < this.numMixtures; d++)
            for (int y = 0; y < this.numVars; y++) {
                int numParents = this.yyParents[y].length + this.yxParents[y].length;
                this.lambda[d][y] = new double[numParents];
            }
        this.invPsi = new double[this.numVars];

        //Other
        this.impliedSecondMoments = new double[this.numMixtures][][];
        this.impliedCovarObserved = new double[this.numMixtures][this.numVars][this.numVars];
        this.invImplCovarObs = new double[this.numMixtures][][];
        this.impliedMu = new double[this.numMixtures][this.numLatents + this.numVars];
        this.detCov = new double[this.numMixtures];
    }

    public void estimate() {
        estimate(true);
    }

    public void estimate(boolean initialize) {
        this.converging = true;
        if (initialize)
            initializeParameters();
        expectation();
        int iter = 0;
        double f, newF = logLikelihood();
        //System.out.println(" ------ F0   = " + newF);
        do {
            f = newF;
            maximization();
            expectation();
            newF = logLikelihood();
            //System.out.println(" ------ F    = " + newF);
            iter++;
        } while (this.converging && newF - f > MIN_TOLERANCE && iter < MAX_ITER);
        if (!this.converging || newF < f) {
            this.score = -Double.MAX_VALUE;
            System.out.println("BOOM!" + this.converging);
            //System.exit(0);
        }
        else
            this.score = newF;
    }

    private void initializeParameters() {
        Random r = new Random();
        //Pi priors
        for (int s = 0; s < this.numMixtures; s++)
            this.pi[s] = 1. / this.numMixtures;

        //Q distribution parameters
        for (int i = 0; i < this.sampleSize; i++) {
            double qSum = .0;
            for (int s = 0; s < this.numMixtures; s++) {
                this.h[s][i] = r.nextDouble() + 0.1;
                //this.h[s][i] = 1.;
                qSum += this.h[s][i];
            }
            for (int s = 0; s < this.numMixtures; s++)
                this.h[s][i] /= qSum;
        }
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++) {
                for (int x = 0; x < this.numLatents; x++)
                    this.eX[d][i][x] = r.nextGaussian();
                for (int x1 = 0; x1 < this.numLatents; x1++)
                    for (int x2 = 0; x2 < this.numLatents; x2++) {
                        if (x1 == x2)
                            this.eXX[d][i][x1][x2] = 1. + this.eX[d][i][x1] * this.eX[d][i][x2];
                        else
                            this.eXX[d][i][x1][x2] = this.eX[d][i][x1] * this.eX[d][i][x2];
                    }
            }
        for (int d = 0; d < this.numMixtures; d++)
            for (int x = 0; x < this.numLatents; x++) {
                for (int p = 0; p < this.xxParents[x].length + this.xyParents[x].length; p++)
                    this.beta[d][x][p] = r.nextGaussian();
                this.upsilon[d][x] = r.nextDouble();
            }
        for (int d = 0; d < this.numMixtures; d++)
            for (int y = 0; y < this.numVars; y++)
                for (int p = 0; p < this.lambda[d][y].length; p++)
                    this.lambda[d][y][p] = r.nextGaussian();
        for (int y = 0; y < this.numVars; y++)
            this.invPsi[y] = r.nextDouble() + 0.2;
   }

    protected void maximization() {
        updateQPi(); //expectation(); System.out.println("          ------ F1    = " + logLikelihood());
        updateQBetaMu(); //expectation(); System.out.println("          ------ F2    = " + logLikelihood());
        updateQUpsilon(); //expectation(); System.out.println("          ------ F3    = " + logLikelihood());
        updateQLambda(); //expectation(); System.out.println("          ------ F4    = " + logLikelihood());
        updateQPsi(); //expectation(); System.out.println("          ------ F5    = " + logLikelihood());
    }

    protected void updateQPi() {
        double totalSum = 0.;
        for (int s = 0; s < this.numMixtures; s++) {
            this.pi[s] = 0.;
            for (int i = 0; i < this.sampleSize; i++)
                this.pi[s] += this.h[s][i];
            totalSum += this.pi[s];
        }
        for (int s = 0; s < this.numMixtures; s++)
            this.pi[s] /= totalSum;
    }

    protected void updateQUpsilon() {
        for (int d = 0; d < this.numMixtures; d++) {
            double sumQS = 0.;
            for (int i = 0; i < this.sampleSize; i++)
                sumQS += this.h[d][i];
            for (int l = 0; l < this.numLatents; l++) {
                double dataContrib = 0;
                for (int i = 0; i < this.sampleSize; i++) {
                    double contr = this.eXX[d][i][l][l];
                    for (int p1 = 0; p1 < this.xxParents[l].length; p1++) {
                        contr -= 2. * this.beta[d][l][p1] *
                                      this.eXX[d][i][l][this.xxParents[l][p1]];
                        for (int p2 = 0; p2 < this.xxParents[l].length; p2++)
                            contr += this.beta[d][l][p1] * this.beta[d][l][p2] *
                                     this.eXX[d][i][this.xxParents[l][p1]][this.xxParents[l][p2]];
                        for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                            contr += 2 * this.beta[d][l][p1] * this.beta[d][l][this.xxParents[l].length + p2] *
                                         this.eX[d][i][this.xxParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                    }
                    for (int p1 = 0; p1 < this.xyParents[l].length; p1++) {
                        contr -= 2. * this.beta[d][l][this.xxParents[l].length + p1] *
                                            this.eX[d][i][l] * this.rawData[i][this.xyParents[l][p1]];
                        for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                            contr += this.beta[d][l][this.xxParents[l].length + p1] * this.beta[d][l][this.xxParents[l].length + p2] *
                                           this.rawData[i][this.xyParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                    }
                    dataContrib += this.h[d][i] * contr;
                }
                this.upsilon[d][l] = sumQS / dataContrib;
            }
        }
    }

    protected void updateQBetaMu() {
        for (int d = 0; d < this.numMixtures; d++)
            for (int l = 0; l < this.numLatents; l++) {
                if (this.xxParents[l].length + this.xyParents[l].length == 0)
                    continue;
                int numParents = this.xxParents[l].length + this.xyParents[l].length;
                double auxS[][] = new double[numParents][numParents];
                double auxMean[] = new double[numParents];
                for (int i = 0; i < auxS.length; i++) {
                    auxMean[i] = 0.;
                    for (int j = 0; j < auxS.length; j++)
                        auxS[i][j] = 0.;
                }
                for (int i = 0; i < this.sampleSize; i++) {
                    double constant = this.h[d][i] * this.upsilon[d][l];
                    for (int p1 = 0; p1 < this.xxParents[l].length; p1++) {
                        for (int p2 = p1; p2 < this.xxParents[l].length; p2++)
                            auxS[p1][p2] += constant *
                                            this.eXX[d][i][this.xxParents[l][p1]][this.xxParents[l][p2]];
                        for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                            auxS[p1][this.xxParents[l].length + p2] += constant *
                                       this.eX[d][i][this.xxParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                        auxMean[p1] += constant * this.eXX[d][i][l][this.xxParents[l][p1]];
                    }
                    for (int p1 = 0; p1 < this.xyParents[l].length; p1++) {
                        for (int p2 = p1; p2 < this.xyParents[l].length; p2++)
                            auxS[this.xxParents[l].length + p1][this.xxParents[l].length + p2] += constant *
                                        this.rawData[i][this.xyParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                        auxMean[this.xxParents[l].length + p1] += constant * this.eX[d][i][l]*
                                                                  this.rawData[i][this.xyParents[l][p1]];
                    }
                }
                for (int p1 = 0; p1 < numParents; p1++)
                    for (int p2 = p1 + 1; p2 < numParents; p2++)
                        auxS[p2][p1] = auxS[p1][p2];
                this.beta[d][l] = MatrixUtils.product(MatrixUtils.inverse(auxS), auxMean);
            }
    }

    protected void updateQLambda() {
        for (int v = 0; v < this.numVars; v++)
            updateQLambda(v);
    }

    protected void updateQLambda(int v) {
        for (int d = 0; d < this.numMixtures; d++) {
            int totalParents = this.yyParents[v].length + this.yxParents[v].length;
            double auxS[][] = new double[totalParents][totalParents];
            double auxMean[] = new double[totalParents];
            for (int i = 0; i < auxS.length; i++) {
                auxMean[i] = 0.;
                for (int j = 0; j < auxS.length; j++)
                    auxS[i][j] = 0.;
            }
            for (int i = 0; i < this.sampleSize; i++) {
                double constant = this.h[d][i];
                for (int p1 = 0; p1 < this.yyParents[v].length; p1++) {
                    for (int p2 = p1; p2 < this.yyParents[v].length; p2++)
                        auxS[p1][p2] += constant *
                                        this.rawData[i][this.yyParents[v][p1]] * this.rawData[i][this.yyParents[v][p2]];
                    for (int p2 = 0; p2 < this.yxParents[v].length; p2++)
                        auxS[p1][this.yyParents[v].length + p2] += constant *
                                   this.rawData[i][this.yyParents[v][p1]] * this.eX[d][i][this.yxParents[v][p2]];
                    auxMean[p1] += constant * this.rawData[i][v] * this.rawData[i][this.yyParents[v][p1]];
                }
                for (int p1 = 0; p1 < this.yxParents[v].length; p1++) {
                    for (int p2 = p1; p2 < this.yxParents[v].length; p2++)
                        auxS[this.yyParents[v].length + p1][this.yyParents[v].length + p2] += constant *
                                this.eXX[d][i][this.yxParents[v][p1]][this.yxParents[v][p2]];
                    auxMean[this.yyParents[v].length + p1] +=
                            constant * this.rawData[i][v] * this.eX[d][i][this.yxParents[v][p1]];
                }
            }
            for (int p1 = 0; p1 < totalParents; p1++)
                for (int p2 = p1 + 1; p2 < totalParents; p2++)
                    auxS[p2][p1] = auxS[p1][p2];
            this.lambda[d][v] = MatrixUtils.product(MatrixUtils.inverse(auxS), auxMean);
        }
    }

    protected void updateQPsi() {
        for (int y = 0; y < this.numVars; y++) {
            double dataContrib = 0.;
            for (int d = 0; d < this.numMixtures; d++) {
                for (int i = 0; i < this.sampleSize; i++) {
                    double partial = this.rawData[i][y];
                    for (int p = 0; p < this.yyParents[y].length; p++)
                        partial -= this.lambda[d][y][p] *
                                   this.rawData[i][this.yyParents[y][p]];
                    for (int p = 0; p < this.yxParents[y].length; p++)
                        partial -=  this.lambda[d][y][this.yyParents[y].length + p] *
                                    this.eX[d][i][this.yxParents[y][p]];
                    dataContrib += this.h[d][i] * this.rawData[i][y] * partial;
                }
            }
            this.invPsi[y] = this.sampleSize / dataContrib;
        }
    }

    protected void singleOptimize(int y) {
        updateQLambda(y);
        double dataContrib = 0.;
        for (int d = 0; d < this.numMixtures; d++) {
            for (int i = 0; i < this.sampleSize; i++) {
                double partial = this.rawData[i][y];
                for (int p = 0; p < this.yyParents[y].length; p++)
                    partial -= this.lambda[d][y][p] *
                               this.rawData[i][this.yyParents[y][p]];
                for (int p = 0; p < this.yxParents[y].length; p++)
                    partial -=  this.lambda[d][y][this.yyParents[y].length + p] *
                                this.eX[d][i][this.yxParents[y][p]];
                dataContrib += this.h[d][i] * this.rawData[i][y] * partial;
            }
        }
        this.invPsi[y] = this.sampleSize / dataContrib;
    }

    protected void iteratedSingleOptimize(int ys[]) {
        for (int c = 0; c < 3; c++) {
            expectation();
            for (int v = 0; v < ys.length; v++) {
                int y = ys[v];
                updateQLambda(y);
                double dataContrib = 0.;
                for (int d = 0; d < this.numMixtures; d++) {
                    for (int i = 0; i < this.sampleSize; i++) {
                        double partial = this.rawData[i][y];
                        for (int p = 0; p < this.yyParents[y].length; p++)
                            partial -= this.lambda[d][y][p] *
                                       this.rawData[i][this.yyParents[y][p]];
                        for (int p = 0; p < this.yxParents[y].length; p++)
                            partial -=  this.lambda[d][y][this.yyParents[y].length + p] *
                                        this.eX[d][i][this.yxParents[y][p]];
                        dataContrib += this.h[d][i] * this.rawData[i][y] * partial;
                    }
                }
                this.invPsi[y] = this.sampleSize / dataContrib;
            }
        }
    }

    protected void momentCalculation() {
        //Build implied covariance matrices
        for (int d = 0; d < this.numMixtures; d++) {
            //Build big matrix of loadings and errors with the required zeroes, stacking latents first, observables last
            double bigLambda[][] = new double[this.numLatents + this.numVars][this.numLatents + this.numVars];
            double bigError[][] =  new double[this.numLatents + this.numVars][this.numLatents + this.numVars];
            for (int x1 = 0; x1 < this.numLatents; x1++) {
                bigError[x1][x1] = 1. / this.upsilon[d][x1];
                for (int x2 = 0; x2 < this.numLatents; x2++) {
                    bigLambda[x1][x2] = 0.;
                    if (x1 != x2)
                        bigError[x1][x2] = 0.;
                }
                for (int y2 = 0; y2 < this.numVars; y2++) {
                    bigLambda[x1][this.numLatents + y2] = 0.;
                    bigError[x1][this.numLatents + y2] = 0.;
                }
            }
            for (int y1 = 0; y1 < this.numVars; y1++) {
                bigError[this.numLatents + y1][this.numLatents + y1] = 1. / this.invPsi[y1];
                for (int y2 = 0; y2 < this.numVars; y2++) {
                    bigLambda[this.numLatents + y1][this.numLatents + y2] = 0.;
                    if (y1 != y2)
                        bigError[this.numLatents + y1][this.numLatents + y2] = 0.;
                }
                for (int x2 = 0; x2 < this.numLatents; x2++) {
                    bigLambda[this.numLatents + y1][x2] = 0.;
                    bigError[this.numLatents + y1][x2] = 0.;
                }
            }
            for (int x1 = 0; x1 < this.numLatents; x1++) {
                for (int x2 = 0; x2 < this.xxParents[x1].length; x2++)
                    bigLambda[x1][this.xxParents[x1][x2]] = this.beta[d][x1][x2];
                for (int y2 = 0; y2 < this.xyParents[x1].length; y2++)
                    bigLambda[x1][this.numLatents + this.xyParents[x1][y2]] = this.beta[d][x1][this.xxParents[x1].length + y2];
            }
            for (int y1 = 0; y1 < this.numVars; y1++) {
                for (int y2 = 0; y2 < this.yyParents[y1].length - 1; y2++)
                    bigLambda[this.numLatents + y1][this.numLatents + this.yyParents[y1][y2]] = this.lambda[d][y1][y2];
                for (int x2 = 0; x2 < this.yxParents[y1].length; x2++)
                    bigLambda[this.numLatents + y1][this.yxParents[y1][x2]] = this.lambda[d][y1][this.yyParents[y1].length + x2];
            }
            double meanParams[] = new double[this.numLatents + this.numVars];
            for (int x = 0; x < this.numLatents; x++)
                meanParams[x] = 0.;
            for (int y = 0; y < this.numVars; y++)
                meanParams[this.numLatents + y] = this.lambda[d][y][this.yyParents[y].length - 1];
            double meanParamsMatrix[][] = MatrixUtils.outerProduct(meanParams, meanParams);
            double iMinusLambda[][] = MatrixUtils.inverse(
                                          MatrixUtils.subtract(
                                                  MatrixUtils.identity(this.numLatents + this.numVars),
                                                  bigLambda));
            this.impliedSecondMoments[d] = MatrixUtils.product(
                                               MatrixUtils.product(iMinusLambda,
                                                        MatrixUtils.sum(bigError, meanParamsMatrix)),
                                               MatrixUtils.transpose(iMinusLambda));


            //Copy relevant matrices
            this.impliedMu[d] = MatrixUtils.product(iMinusLambda, meanParams);
            for (int y1 = 0; y1 < this.numVars; y1++)
                for (int y2 = 0; y2 < this.numVars; y2++)
                    this.impliedCovarObserved[d][y1][y2] = this.impliedSecondMoments[d][this.numLatents + y1][this.numLatents + y2] -
                                                           this.impliedMu[d][this.numLatents + y1] *
                                                           this.impliedMu[d][this.numLatents + y2];
            this.invImplCovarObs[d] = MatrixUtils.inverse(this.impliedCovarObserved[d]);
            this.detCov[d] = MatrixUtils.determinant(this.impliedCovarObserved[d]);
        }
    }

    protected void expectation() {
        momentCalculation();
        //Calculate h expectations
        double center[] = new double[this.numVars];
        for (int i = 0; i < this.sampleSize; i++) {
            double sumH = 0.;
            for (int d = 0; d < this.numMixtures; d++) {
                for (int v = 0; v < this.numVars; v++)
                     center[v] = this.rawData[i][v] - this.impliedMu[d][this.numLatents + v];
                this.h[d][i] = this.pi[d] * normalPdf(center, this.invImplCovarObs[d], this.detCov[d]);
                sumH += this.h[d][i];
            }
            for (int d = 0; d < this.numMixtures; d++)
                this.h[d][i] /= sumH;
        }
        //Regress latents on the observed variables to get their new expectations
        if (this.numLatents == 0)
            return;
        for (int d = 0; d < this.numMixtures; d++) {
            double impliedSecondYY_S[][] = new double[this.numVars + 1][this.numVars + 1];
            for (int y1 = 0; y1 < this.numVars; y1++) {
                for (int y2 = 0; y2 < this.numVars; y2++)
                    impliedSecondYY_S[y1][y2] = this.impliedSecondMoments[d][this.numLatents + y1][this.numLatents + y2];
                impliedSecondYY_S[y1][this.numVars] = impliedSecondYY_S[this.numVars][y1] =
                                                      this.impliedMu[d][this.numLatents + y1];
            }
            impliedSecondYY_S[this.numVars][this.numVars] = 1.;

            double impliedSecondXY_S[][] = new double[this.numLatents][this.numVars + 1];
            double impliedSecondXX[][] = new double[this.numLatents][this.numLatents];
            for (int x1 = 0; x1 < this.numLatents; x1++) {
                for (int y2 = 0; y2 < this.numVars; y2++)
                    impliedSecondXY_S[x1][y2] = this.impliedSecondMoments[d][x1][this.numLatents + y2];
                impliedSecondXY_S[x1][this.numVars] = this.impliedMu[d][x1];
                for (int x2 = 0; x2 < this.numLatents; x2++)
                    impliedSecondXX[x1][x2] = this.impliedSecondMoments[d][x1][x2];
            }

            double regression[][] = MatrixUtils.product(impliedSecondXY_S,
                                        MatrixUtils.inverse(impliedSecondYY_S));
            double varZZMatrix[][] = MatrixUtils.subtract(impliedSecondXX,
                                       MatrixUtils.product(impliedSecondXY_S, MatrixUtils.transpose(regression)));

            for (int x = 0; x < this.numLatents; x++)
                for (int i = 0; i < this.sampleSize; i++) {
                    this.eX[d][i][x] = regression[x][this.numVars];
                    for (int y = 0; y < this.numVars; y++)
                        this.eX[d][i][x] += regression[x][y] * this.rawData[i][y];
                }
            for (int x1 = 0; x1 < this.numLatents; x1++)
                for (int x2 = x1; x2 < this.numLatents; x2++)
                    for (int i = 0; i < this.sampleSize; i++)
                        this.eXX[d][i][x1][x2] = this.eXX[d][i][x2][x1] =
                                varZZMatrix[x1][x2] + this.eX[d][i][x1] * this.eX[d][i][x2];
        }
    }

    /******************************************************************************************************************
     * Score function evaluation
     */

    protected double logLikelihood(double rawData[][]) {
        int sampleSize = rawData.length;
        double result = 0.;
        double center[] = new double[this.numVars];
        for (int i = 0; i < sampleSize; i++) {
            double compSum = 0.;
            for (int d = 0; d < this.numMixtures; d++) {
                for (int v = 0; v < this.numVars; v++)
                    center[v] = rawData[i][v] - this.impliedMu[d][this.numLatents + v];
                compSum += normalPdf(center, this.invImplCovarObs[d], this.detCov[d]) * this.pi[d];
            }
            if (compSum > 0.)
                result += Math.log(compSum);
            else {
                //result += Math.log(Double.MIN_VALUE);
                this.converging = false;
                return Double.MAX_VALUE;
            }
            if (Double.isInfinite(result)) {
                System.out.println("To the infinity and beyond...");
                this.converging = false;
                return Double.MAX_VALUE;
            }
        }
        return result;
    }

    public double logLikelihood() {
        return logLikelihood(this.rawData);
    }

    public double logLikelihoodEval(RectangularDataSet external) {
        double raw[][] = new double[external.getNumRows()][this.numVars + 1];
        for (int i = 0; i < this.numVars; i++) {
            Node v = external.getVariable(this.indicators[i].toString());
            double raw0[] = external.getDoubleMatrix().viewColumn(external.getColumn(v)).toArray();
            for (int j = 0; j < raw0.length; j++)
                raw[j][i] = raw0[j];
        }
        for (int j = 0; j < external.getNumRows(); j++)
            raw[j][this.numVars] = 1.;
        return logLikelihood(raw);
    }

    private double normalPdf(double center[], double invCovar[][], double det) {
        return  1. / (Math.sqrt(Math.pow((2 * Math.PI), center.length)) * Math.sqrt(det)) *
                Math.exp(-0.5 * MatrixUtils.innerProduct(center, MatrixUtils.product(invCovar, center)));
    }

    //
    // STRUCTURAL EM METHODS AND FIELDS FOR PURIFICATION ALGORITHMS
    //

    public double commitFastNewFreeVariableScoring(String variable) {
        return fastNewFreeVariableScoring(variable, true);
    }

    public double fastNewFreeVariableScoring(String variable) {
        return fastNewFreeVariableScoring(variable, false);
    }

    protected double fastNewFreeVariableScoring(String variable, boolean commit) {
        int y = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable)) {
                y = i;
                break;
            }
        if (y < 0)
            throw new RuntimeException("This variable does not exist: " + variable);

        Node nodeY = this.graph.getNode(this.indicators[y].toString());
        if (this.graph.getChildren(nodeY).size() > 0)
            throw new RuntimeException("This variable should not have any children: " + variable);

        int yyParentsBackup[] = this.yyParents[y], yxParentsBackup[] = this.yxParents[y];
        double invPsiBackup = this.invPsi[y];
        double lambdaBackup[][] = new double[this.numMixtures][];
        for (int d = 0; d < this.numMixtures; d++)
            lambdaBackup[d] = this.lambda[d][y];
        this.yyParents[y] = new int[this.indicators.length];
        int niCount = 0;
        for (int i = 0; i < this.indicators.length; i++) {
            if (i == y)
                continue;
            this.yyParents[y][niCount++] = i;
        }
        this.yyParents[y][this.indicators.length - 1] = this.numVars;
        this.yxParents[y] = new int[0];
        for (int d = 0; d < this.numMixtures; d++)
            this.lambda[d][y] = new double[this.yyParents[y].length + this.yxParents[y].length];
        expectation();
        singleOptimize(y);
        momentCalculation();
        double newScore = logLikelihood() - 0.5 *
                          (this.numParameters + (this.yyParents[y].length - yyParentsBackup.length) * this.numMixtures) *
                          Math.log(this.sampleSize);

        if (!commit) {
            this.yyParents[y] = yyParentsBackup;
            this.yxParents[y] = yxParentsBackup;
            this.invPsi[y] = invPsiBackup;
            for (int d = 0; d < this.numMixtures; d++)
                this.lambda[d][y] = lambdaBackup[d];
            expectation();
        } else
            pack(y);
        return newScore;
    }

    public double slowNewFreeVariableScoring(String variable) {
        int y = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable)) {
                y = i;
                break;
            }
        if (y < 0)
            throw new RuntimeException("This variable does not exist: " + variable);
        Node nodeY = this.graph.getNode(this.indicators[y].toString());
        if (this.graph.getChildren(nodeY).size() > 0)
            throw new RuntimeException("This variable should not have any children: " + variable);

        this.yyParents[y] = new int[this.indicators.length];
        int niCount = 0;
        for (int i = 0; i < this.indicators.length; i++) {
            if (i == y)
                continue;
            this.yyParents[y][niCount++] = i;
        }
        this.yyParents[y][this.indicators.length - 1] = this.numVars;
        this.yxParents[y] = new int[0];
        Random r = new Random();
        for (int d = 0; d < this.numMixtures; d++) {
            this.lambda[d][y] = new double[this.yyParents[y].length + this.yxParents[y].length];
            for (int i = 0; i < this.yyParents[y].length + this.yxParents[y].length; i++)
                this.lambda[d][y][i] = r.nextGaussian();
        }
        estimate(false);
        return logLikelihood();
    }

    public void pack(String variable) {
        int y = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable)) {
                y = i;
                break;
            }
        if (y < 0)
            throw new RuntimeException("This variable does not exist: " + variable);
        pack(y);
    }

    public void pack(int y) {
        this.graph.removeNode(this.graph.getNode(this.indicators[y].toString()));
        if (this.originalGraph.getNode(this.indicators[y].toString()) != null)
            this.originalGraph.removeNode(this.originalGraph.getNode(this.indicators[y].toString()));
        this.numVars = this.indicators.length - 1;
        Node   newIndicators[] = new Node[this.numVars];
        int    newYXParents[][] = new int[this.numVars][];
        double newInvPsi[] = new double[this.numVars];
        int newIndices[] = new int[this.indicators.length];
        int niCount = 0;
        for (int i = 0; i < this.indicators.length; i++) {
            if (i != y) {
                newIndicators[niCount] = this.indicators[i];
                newYXParents[niCount] = this.yxParents[i];
                newInvPsi[niCount] = this.invPsi[i];
                newIndices[i] = niCount;
                niCount++;
            }
        }
        niCount = 0;
        int newYYParents[][] = new int[this.numVars][];
        for (int i = 0; i < this.indicators.length; i++) {
            if (i != y) {
                newYYParents[niCount] = new int[this.yyParents[i].length];
                for (int p = 0; p < this.yyParents[i].length - 1; p++)
                     newYYParents[niCount][p] = newIndices[this.yyParents[i][p]];
                newYYParents[niCount][this.yyParents[i].length - 1] = this.numVars;
                niCount++;
            }
        }
        double newLambda[][][] = new double[this.numMixtures][this.numVars][];
        niCount = 0;
        for (int i = 0; i < this.indicators.length; i++)
            if (i != y) {
                for (int d = 0; d < this.numMixtures; d++)
                    newLambda[d][niCount] = this.lambda[d][i];
                niCount++;
            }
        this.indicators = newIndicators;
        this.yxParents = newYXParents;
        this.yyParents = newYYParents;
        this.invPsi = newInvPsi;
        for (int i = 0; i < this.numVars; i++)
            for (int d = 0; d < this.numMixtures; d++)
                this.lambda[d][i] = newLambda[d][i];
        this.rawData = new double[this.sampleSize][this.numVars + 1];
        double matrix[][] = this.dataSet.getDoubleMatrix().toArray();
        for (int i = 0; i < this.numVars; i++)
            for (int j = 0; j < this.sampleSize; j++)
                this.rawData[j][i] = matrix[j][i];
        for (int j = 0; j < this.sampleSize; j++)
            this.rawData[j][this.numVars] = 1.;
        this.impliedCovarObserved = new double[this.numMixtures][this.numVars][this.numVars];
        this.impliedMu = new double[this.numMixtures][this.numLatents + this.numVars];
        momentCalculation(); expectation();
        this.score = logLikelihood();
    }

    public Object clone() {
        MixtureOfSemsEstimator newSem = new MixtureOfSemsEstimator(this.dataSet, this.graph, this.numMixtures);
        for (int d = 0; d < this.numMixtures; d++) {
            newSem.pi[d] = this.pi[d];
            for (int i = 0; i < this.sampleSize; i++) {
                newSem.h[d][i] = this.h[d][i];
                for (int x1 = 0; x1 < this.numLatents; x1++) {
                    newSem.eX[d][i][x1] = this.eX[d][i][x1];
                    for (int p = 0; p < this.beta[d][x1].length; p++)
                        newSem.beta[d][x1][p] = this.beta[d][x1][p];
                    for (int x2 = 0; x2 < this.numLatents; x2++)
                        newSem.eXX[d][i][x1][x2] = this.eXX[d][i][x1][x2];
                    newSem.upsilon[d][x1] = this.upsilon[d][x1];
                }
                for (int y1 = 0; y1 < this.numVars; y1++) {
                    for (int p = 0; p < this.lambda[d][y1].length; p++)
                        newSem.lambda[d][y1][p] = this.lambda[d][y1][p];
                    newSem.invPsi[y1] = this.invPsi[y1];
                }
            }
        }
        newSem.momentCalculation();
        newSem.score = this.score;
        return newSem;
    }

    public MixtureOfSemsEstimator getMarginal(String variable) {
        MixtureOfSemsEstimator marginal = (MixtureOfSemsEstimator) clone();
        int y = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable)) {
                y = i;
                break;
            }
        if (y < 0)
            throw new RuntimeException("This variable does not exist: " + variable);
        marginal.pack(y);
        return marginal;
    }

    public double getScore() {
        //return logLikelihood() - 0.5 * this.numParameters * Math.log(this.sampleSize);
        return this.score;
    }

    public double fastBidirectedEdgeAddition(String variable1, String variable2) {
        int y[] = new int[2];
        y[0] = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable1)) {
                y[0] = i;
                break;
            }
        if (y[0] < 0)
            throw new RuntimeException("This variable does not exist: " + variable1);

        y[1] = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable1)) {
                y[1] = i;
                break;
            }
        if (y[1] < 0)
            throw new RuntimeException("This variable does not exist: " + variable2);

        Node latentsBackup[] = new Node[this.latents.length + 1];
        for (int i = 0; i < this.latents.length; i++) {
            latentsBackup[i] = this.latents[i];
        }
        latentsBackup[this.latents.length] = new GraphNode("_L" + this.graph.getNodes().size());
        latentsBackup[this.latents.length].setNodeType(NodeType.LATENT);
        double exB[][][] = new double[this.numMixtures][this.sampleSize][this.numLatents + 1];
        double exxB[][][][] = new double[this.numMixtures][this.sampleSize][this.numLatents + 1][this.numLatents + 1];
        double betaB[][][] = new double[this.numMixtures][this.numLatents + 1][];
        double upsilonB[][] =new double[this.numMixtures][this.numLatents + 1];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int i = 0; i < this.sampleSize; i++) {
                for (int x1 = 0; x1 < this.numLatents; x1++) {
                    exB[d][i][x1] = this.eX[d][i][x1];
                    exxB[d][i][this.numLatents][x1] = exxB[d][i][x1][this.numLatents] = 0.;
                    for (int x2 = 0; x2 < this.numLatents; x2++)
                        exxB[d][i][x1][x2] = this.eXX[d][i][x1][x2];
                }
                exB[d][i][this.numLatents] = 0.;
            }
            for (int x1 = 0; x1 < this.numLatents; x1++) {
                betaB[d][x1] = this.beta[d][x1];
                upsilonB[d][x1] = this.upsilon[d][x1];
            }
            betaB[d][this.numLatents] = new double[0];
            upsilonB[d][this.numLatents] = 1.;
        }

        this.eX = exB; this.eXX = exxB; this.beta = betaB; this.upsilon = upsilonB;
        this.graph.addNode(latentsBackup[this.latents.length]);
        this.graph.addDirectedEdge(latentsBackup[this.latents.length], this.graph.getNode(variable1));
        this.graph.addDirectedEdge(latentsBackup[this.latents.length], this.graph.getNode(variable2));
        this.latents = latentsBackup;
        this.numParameters += 2 * this.numMixtures;
        this.numLatents++;

        Random r = new Random();
        for (int i = 0; i < 2; i++) {
            for (int d = 0; d < this.numMixtures; d++) {
                double newLambda[] = new double[this.yyParents[y[i]].length + this.yxParents[y[i]].length + 1];
                for (int j = 0; j < this.yyParents[y[i]].length + this.yxParents[y[i]].length; j++)
                    newLambda[j] = this.lambda[d][y[i]][j];
                newLambda[this.yyParents[y[i]].length + this.yxParents[y[i]].length] = r.nextDouble();
                this.lambda[d][y[i]] = newLambda;
            }
            int newYXParents[] = new int[this.yxParents[y[i]].length + 1];
            for (int j = 0; j < this.yxParents[y[i]].length; j++)
                newYXParents[j] = this.yxParents[y[i]][j];
            newYXParents[this.yxParents[y[i]].length] = this.numLatents - 1;
            this.yxParents[y[i]] = newYXParents;
        }
        this.impliedSecondMoments = new double[this.numMixtures][this.numLatents][this.numLatents];

        int xxPB[][] = new int[this.numLatents][];
        int xyPB[][] = new int[this.numLatents][];
        for (int i = 0; i < this.numLatents -1; i++) {
            xxPB[i] = this.xxParents[i];
            xyPB[i] = this.xyParents[i];
        }
        this.xxParents = xxPB; this.xyParents = xyPB;
        this.xxParents[this.numLatents - 1] = new int[0];
        this.xyParents[this.numLatents - 1] = new int[0];
        iteratedSingleOptimize(y);
        momentCalculation();
        return logLikelihood();
    }

    public double fastNewSingleEdgeScoring(String parent, String variable) {
        int y = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(variable)) {
                y = i;
                break;
            }
        if (y < 0)
            throw new RuntimeException("This variable does not exist: " + variable);

        int p = -1;
        for (int i = 0; i < this.numVars; i++)
            if (this.indicators[i].toString().equals(parent)) {
                p = i;
                break;
            }
        if (p < 0)
            throw new RuntimeException("This variable does not exist: " + parent);

        Node nodeY = this.graph.getNode(this.indicators[y].toString());
        if (this.graph.getChildren(nodeY).size() > 0)
            throw new RuntimeException("This variable should not have any children: " + variable);

        int yyParentsBackup[] = this.yyParents[y];
        double invPsiBackup = this.invPsi[y];
        double lambdaBackup[][] = new double[this.numMixtures][];
        for (int d = 0; d < this.numMixtures; d++)
            lambdaBackup[d] = this.lambda[d][y];
        this.yyParents[y] = new int[2];
        this.yyParents[y][0] = p; this.yyParents[y][1] = this.numVars;
        for (int d = 0; d < this.numMixtures; d++)
            this.lambda[d][y] = new double[this.yyParents[y].length + this.yxParents[y].length];
        expectation();
        singleOptimize(y);
        momentCalculation();
        double newScore = logLikelihood() - 0.5 *
                          (this.numParameters + (this.yyParents[y].length - yyParentsBackup.length) * this.numMixtures) *
                          Math.log(this.sampleSize);

        this.yyParents[y] = yyParentsBackup;
        this.invPsi[y] = invPsiBackup;
        for (int d = 0; d < this.numMixtures; d++)
            this.lambda[d][y] = lambdaBackup[d];
        expectation();

        return newScore;
    }

    public void printImpliedCovar() {
        for (int y1 = 0; y1 < this.numVars; y1++) {
            for (int y2 = 0; y2 < this.numVars; y2++)
                System.out.print((((int) (1000 * this.impliedCovarObserved[0][y1][y2]))) / 1000. + "\t");
            System.out.println();
       }
    }

    public void printLambdas() {
        for (int y1 = 0; y1 < this.numVars; y1++) {
            for (int y2 = 0; y2 < this.lambda[0][y1].length; y2++)
                System.out.print("L[0][" + y1 + "][" + y2 + "] = " + (((int) (1000 * this.lambda[0][y1][y2]))) / 1000. + "\t");
            System.out.println();
       }
       for (int i = 0; i < this.numLatents; i++)
            System.out.println("U[" + 0 + "][" + i + "] = " + this.upsilon[0][i]);
    }
}


